import { parseArgs } from "node:util";
import chalk from "chalk";
import { version } from "../package.json" with { type: "json" };

export const USAGE = `
${chalk.bold("muse")} - Compile and process troves of data
${chalk.dim(`v${version}`)}

Usage:
  muse [/path/to/binding.yaml] <options>

Options:
  --stdout    -s  Output final data to stdout
  --verbose,  -v  Enable verbose logging
  --help,     -h  Show this help message
`.trim();

/**
 * A shape representing the arguments passed to the CLI
 */
export interface CLIArguments {
  inputFilePath: string;

  options: {
    help: boolean;
    verbose: boolean;
    stdout: boolean;
  };
}

/**
 * Given an array of CLI arguments, parse them into a structured object
 *
 * @param argv The arguments to parse
 * @returns The parsed CLI arguments
 *
 * @throws {CLIError} if the arguments are invalid
 */
export function parseCLIArguments(argv: string[]): CLIArguments {
  const { values: options, positionals: args } = parseArgs({
    args: argv,
    options: {
      help: {
        short: "h",
        default: false,
        type: "boolean",
      },
      verbose: {
        short: "v",
        default: false,
        type: "boolean",
      },
      stdout: {
        short: "s",
        default: false,
        type: "boolean",
      },
    },
    strict: true,
    allowPositionals: true,
  });

  return {
    inputFilePath: args[0] ?? "./binding.yaml",

    options: {
      help: options.help,
      verbose: options.verbose,
      stdout: options.stdout,
    },
  };
}
