import path, { dirname } from "node:path";
import { Glob } from "bun";
import z from "zod";
import { FileNotFoundError, MuseError } from "./errors";
import { parseMuseFile, type MuseEntry } from "./muse-file";

// binding.yaml file schema
const BindingSchema = z.object({
  include: z.array(z.string()).default([]),
  markdownKey: z.string().default("content"),
  files: z
    .array(z.string())
    .default(["**/*.yaml", "**/*.md", "**/*.toml", "**/*.json"]),
  options: z.record(z.string(), z.any()).default({}),
  processors: z.array(z.string()).default([]),
});

type ParsedBinding = z.infer<typeof BindingSchema>;

/**
 * Given a path, find the binding.yaml file
 * If the path is to a directory, check for a binding.yaml inside
 * If the path is to a file, check if it exists and is a binding.yaml
 * Otherwise throw an error
 *
 * @param bindingPath - The path to the binding file
 * @returns The absolute resolved path to the binding file
 */
export async function resolveBindingPath(bindingPath: string): Promise<string> {
  // If the path does not specify a filename, use binding.yaml
  const inputLocation = Bun.file(bindingPath);

  // If it is a directory, try again seeking a binding.yaml inside
  const stat = await inputLocation.stat();
  if (stat.isDirectory()) {
    return resolveBindingPath(path.resolve(bindingPath, "binding.yaml"));
  }

  // If it doesnt exist, bail
  if (!(await inputLocation.exists())) {
    throw new FileNotFoundError(bindingPath);
  }

  // Resolve the path
  return path.resolve(bindingPath);
}

export type MuseProcessorFn = (
  binding: Binding,
  opts: Record<string, unknown>,
) => Promise<Binding>;

export interface MuseProcessor {
  readonly filePath: string;
  readonly name: string;
  readonly description: string;
  readonly process: MuseProcessorFn;
}

export interface Binding {
  readonly _raw: ParsedBinding;
  readonly bindingPath: string;
  readonly markdownKey: string;
  readonly includedFiles: string[];
  readonly entries: MuseEntry[];
  readonly meta: Record<string, unknown>;
  readonly options: Record<string, unknown>;
  readonly processors: MuseProcessor[];
  readonly imports: Binding[];
}

export async function loadBinding(initialPath: string): Promise<Binding> {
  // Resolve the file location if possible
  const bindingPath = await resolveBindingPath(initialPath);
  const bindingDirname = dirname(bindingPath);
  const loadedBindingData = await parseMuseFile(bindingPath);

  if (loadedBindingData.length === 0) {
    throw new MuseError("No entries found in binding file");
  }

  // Load and parse the Binding YAML content
  const parsedBinding = BindingSchema.parse(loadedBindingData[0].data);

  // Prepare collections to load into
  const extendsFrom: Binding[] = [];
  const entries: MuseEntry[] = [];

  // Process imported bindings and bring in their entries
  for (const includePath of parsedBinding.include) {
    const pathFromHere = path.resolve(bindingDirname, includePath);
    const extendBindingPath = await resolveBindingPath(pathFromHere);
    const loadedBinding = await loadBinding(extendBindingPath);
    extendsFrom.push(loadedBinding);
    entries.push(...loadedBinding.entries);
  }

  // Collect file manifest from globs
  const allFilePaths: string[] = [];
  for (const thisGlob of parsedBinding.files) {
    const glob = new Glob(thisGlob);

    const results = glob.scanSync({
      cwd: bindingDirname,
      absolute: true,
      followSymlinks: true,
      onlyFiles: true,
    });

    allFilePaths.push(...results);
  }

  // Exclude the binding.yaml file
  const includedFilePaths = allFilePaths.filter((filePath) => {
    return filePath !== bindingPath;
  });

  const fileLoadPromises: Promise<MuseEntry[]>[] = [];
  for (const filePath of includedFilePaths) {
    fileLoadPromises.push(
      parseMuseFile(filePath, {
        markdownKey: parsedBinding.markdownKey,
      }),
    );
  }

  const mainEntries = await Promise.all(fileLoadPromises);
  entries.push(...mainEntries.flat());

  // Load and check processors
  const processors: MuseProcessor[] = [];
  for (const processorPath of parsedBinding.processors) {
    const resolvedProcessorPath = path.resolve(bindingDirname, processorPath);
    const { process, name, description } = await import(resolvedProcessorPath);

    if (!process || typeof process !== "function") {
      throw new MuseError(
        `Processor at ${processorPath} does not export a process() function`,
      );
    }

    const processorName = String(name ?? "Unnamed Processor");
    const processorDescription = String(
      description ?? "No description provided",
    );

    processors.push({
      filePath: resolvedProcessorPath,
      process: process as MuseProcessor["process"],
      name: processorName,
      description: processorDescription,
    });
  }

  return {
    _raw: parsedBinding,
    bindingPath,
    markdownKey: parsedBinding.markdownKey,
    includedFiles: includedFilePaths,
    entries,
    options: parsedBinding.options,
    processors: processors,
    imports: extendsFrom,
    meta: {},
  };
}
